@@ -8,7 +8,7 @@ use rocfl::{cmd, config};
 use structopt::StructOpt;
 
 fn main() {
-    let args = RocflArgs::from_args();
+    let mut args = RocflArgs::from_args();
 
     let log_level = if args.quiet {
         LevelFilter::Off
@@ -35,6 +35,11 @@ fn main() {
         }
     };
 
+    // If the output is being piped then we should disable styling
+    if atty::isnt(atty::Stream::Stdout) {
+        args.no_styles = true;
+    }
+
     if let Err(e) = cmd::exec_command(&args, config) {
         match e {
             RocflError::CopyMoveError(errors) => {
