@@ -72,7 +72,7 @@ pub struct RocflArgs {
     #[structopt(short, long, value_name = "PROFILE")]
     pub profile: Option<String>,
 
-    /// Suppress error messages
+    /// Suppress error messages and other command specific logging
     #[structopt(short, long)]
     pub quiet: bool,
 
@@ -122,6 +122,8 @@ pub enum Command {
     Status(StatusCmd),
     #[structopt(name = "purge")]
     Purge(PurgeCmd),
+    #[structopt(name = "validate")]
+    Validate(ValidateCmd),
 }
 
 /// Edit rocfl configuration
@@ -141,7 +143,9 @@ pub struct ConfigCmd {}
 /// List objects or files within objects
 ///
 /// When listing objects, rocfl must scan the entire repository, and can therefore be very slow
-/// when operating on large repositories or repositories in S3.
+/// when operating on large repositories or repositories in S3. Results will be printed as soon
+/// as they're found so long as the results do not need to be sorted or displayed in a formatted
+/// table. Use '-t' to disable the table formatting.
 ///
 /// This command supports glob expressions. When you use globs, it is usually a good idea to
 /// quote them so that your shell does not attempt to expand them.
@@ -186,11 +190,11 @@ pub struct ListCmd {
     #[structopt(short, long, value_name = "VERSION")]
     pub version: Option<VersionNum>,
 
-    /// Field to sort on
+    /// Field to sort on. By default, objects are unsorted and object contents are sorted on name.
     #[structopt(short, long,
     value_name = "FIELD",
     possible_values = &Field::variants(),
-    default_value = "Name",
+    default_value = "Default",
     case_insensitive = true)]
     pub sort: Field,
 
@@ -542,14 +546,42 @@ pub struct PurgeCmd {
     pub object_id: String,
 }
 
+/// Validate an object or the entire repository
+///
+/// When run on a specific object, the object is validated against the OCFL spec, and any issues
+/// are reported. When run against the entire repository, the repository structure is validated,
+/// in addition to validating all of the objects in the repository.
+///
+/// Return code 1 is returned if there were problems performing the validation, but no invalid
+/// objects were identified. Return code 2 is returned if invalid objects were identified. Return
+/// code 0 is returned if all objects were valid, or only warning level issues were identified.
+///
+/// Use '--quiet' to suppress output for valid objects. For example: 'rocfl --quiet validate'
+#[derive(Debug, StructOpt)]
+#[structopt(setting(ColorAuto), setting(ColoredHelp), setting(DisableVersion))]
+pub struct ValidateCmd {
+    /// Interpret positional parameters as paths to object roots relative the repository root
+    #[structopt(short, long)]
+    pub paths: bool,
+
+    /// Disable fixity check on stored files
+    #[structopt(short, long)]
+    pub no_fixity_check: bool,
+
+    /// IDs of the objects to validate, or paths object roots when used with '--paths'
+    #[structopt(name = "OBJ_ID/PATH")]
+    pub object_ids: Vec<String>,
+}
+
 // TODO a command for rebasing staging if an object is updated after the staged version was created?
 
 #[derive(Debug)]
 pub struct Num(pub usize);
 
 arg_enum! {
-    #[derive(Debug, Clone, Copy)]
+    #[derive(Debug, Clone, Copy, Eq, PartialEq)]
     pub enum Field {
+        Default,
         Name,
         Version,
         Updated,
