@@ -9,7 +9,7 @@ use std::rc::Rc;
 use std::sync::atomic::{AtomicBool, Ordering};
 
 use chrono::Local;
-use log::{error, info, warn};
+use log::{info, warn};
 use once_cell::sync::OnceCell;
 #[cfg(feature = "s3")]
 use rusoto_core::Region;
@@ -25,15 +25,16 @@ use crate::ocfl::store::layout::{LayoutExtensionName, StorageLayout};
 #[cfg(feature = "s3")]
 use crate::ocfl::store::s3::S3OcflStore;
 use crate::ocfl::store::{OcflStore, StagingStore};
+use crate::ocfl::validate::ObjectValidationResult;
 use crate::ocfl::{
-    paths, util, CommitMeta, Diff, DigestAlgorithm, InventoryPath, ObjectVersion,
-    ObjectVersionDetails, VersionDetails, VersionNum,
+    paths, util, validate, CommitMeta, ContentPath, Diff, DigestAlgorithm, IncrementalValidator,
+    InventoryPath, LogicalPath, ObjectVersion, ObjectVersionDetails, VersionDetails, VersionNum,
+    VersionRef,
 };
 
-/// Interface for interacting with an OCFL repository
+/// OCFL repository
 pub struct OcflRepo {
-    /// For local filesystem repos, this is the storage root. TBD for S3.
-    // TODO experiment changing this to a generic
+    /// Interfaces with the files on disk or in S3
     store: Box<dyn OcflStore + Sync + Send>,
     /// The OCFL repo that stores staged objects
     staging: OnceCell<FsOcflStore>,
@@ -95,6 +96,7 @@ impl OcflRepo {
         region: Region,
         bucket: &str,
         prefix: Option<&str>,
+        // TODO fix the AsRef<Path> stuff to call inner methods -- I might just wait for https://github.com/rust-lang/rust/issues/77960
         staging_root: impl AsRef<Path>,
         layout: Option<StorageLayout>,
         profile: Option<&str>,
@@ -137,22 +139,58 @@ impl OcflRepo {
         self.store.close();
     }
 
+    /// Validates the specified object and returns any problems found. Err will only be returned
+    /// if a non-validation problem was encountered.
+    pub fn validate_object(
+        &self,
+        object_id: &str,
+        fixity_check: bool,
+    ) -> Result<ObjectValidationResult> {
+        self.ensure_open()?;
+        self.store.validate_object(object_id, fixity_check)
+    }
+
+    /// Validates the specified object at the specified path, relative the storage root, and
+    /// returns any problems found. Err will only be returned if a non-validation problem was
+    /// encountered.
+    pub fn validate_object_at(
+        &self,
+        path: &str,
+        fixity_check: bool,
+    ) -> Result<ObjectValidationResult> {
+        self.ensure_open()?;
+        self.store.validate_object_at(path, fixity_check)
+    }
+
+    /// Validates the structure of an OCFL repository as well as all of the objects in the repository
+    /// When `fixity_check` is `false`, then the digests of object content files are not validated.
+    ///
+    /// The storage root is validated immediately, and an incremental validator is returned that
+    /// is used to lazily validate the rest of the repository.
+    pub fn validate_repo<'a>(
+        &'a self,
+        fixity_check: bool,
+    ) -> Result<Box<dyn IncrementalValidator + 'a>> {
+        self.ensure_open()?;
+        self.store.validate_repo(fixity_check)
+    }
+
     /// Returns an iterator that iterate through all of the objects in an OCFL repository.
     /// Objects are lazy-loaded. An optional glob pattern may be provided to filter the objects
     /// that are returned.
     ///
-    /// The iterator return an error if it encounters a problem accessing an object. This does
+    /// The iterator returns an error if it encounters a problem accessing an object. This does
     /// terminate the iterator; there are still more objects until it returns `None`.
     pub fn list_objects<'a>(
         &'a self,
         filter_glob: Option<&str>,
-    ) -> Result<Box<dyn Iterator<Item = ObjectVersionDetails> + 'a>> {
+    ) -> Result<Box<dyn Iterator<Item = Result<ObjectVersionDetails>> + 'a>> {
         self.ensure_open()?;
 
         let inv_iter = self.store.iter_inventories(filter_glob)?;
 
         Ok(Box::new(InventoryAdapterIter::new(inv_iter, |inventory| {
-            ObjectVersionDetails::from_inventory(inventory, None)
+            ObjectVersionDetails::from_inventory(inventory, VersionRef::Head)
         })))
     }
 
@@ -160,12 +198,12 @@ impl OcflRepo {
     /// Objects are lazy-loaded. An optional glob pattern may be provided to filter the objects
     /// that are returned.
     ///
-    /// The iterator return an error if it encounters a problem accessing an object. This does
+    /// The iterator returns an error if it encounters a problem accessing an object. This does
     /// terminate the iterator; there are still more objects until it returns `None`.
     pub fn list_staged_objects<'a>(
         &'a self,
         filter_glob: Option<&str>,
-    ) -> Result<Box<dyn Iterator<Item = ObjectVersionDetails> + 'a>> {
+    ) -> Result<Box<dyn Iterator<Item = Result<ObjectVersionDetails>> + 'a>> {
         self.ensure_open()?;
 
         if !self.staging_root.exists() {
@@ -175,7 +213,7 @@ impl OcflRepo {
         let inv_iter = self.get_staging()?.iter_inventories(filter_glob)?;
 
         Ok(Box::new(InventoryAdapterIter::new(inv_iter, |inventory| {
-            ObjectVersionDetails::from_inventory(inventory, None)
+            ObjectVersionDetails::from_inventory(inventory, VersionRef::Head)
         })))
     }
 
@@ -184,11 +222,7 @@ impl OcflRepo {
     ///
     /// If the object or version of the object cannot be found, then a `RocflError::NotFound`
     /// error is returned.
-    pub fn get_object(
-        &self,
-        object_id: &str,
-        version_num: Option<VersionNum>,
-    ) -> Result<ObjectVersion> {
+    pub fn get_object(&self, object_id: &str, version_num: VersionRef) -> Result<ObjectVersion> {
         self.ensure_open()?;
 
         let inventory = self.store.get_inventory(object_id)?;
@@ -228,7 +262,7 @@ impl OcflRepo {
 
         ObjectVersion::from_inventory(
             staging_inventory,
-            Some(version),
+            version.into(),
             &root,
             staging.as_ref(),
             util::BACKSLASH_SEPARATOR,
@@ -244,7 +278,7 @@ impl OcflRepo {
     pub fn get_object_details(
         &self,
         object_id: &str,
-        version_num: Option<VersionNum>,
+        version_num: VersionRef,
     ) -> Result<ObjectVersionDetails> {
         self.ensure_open()?;
 
@@ -261,7 +295,7 @@ impl OcflRepo {
 
         let inventory = self.get_staged_inventory(object_id)?;
         let version = inventory.head;
-        ObjectVersionDetails::from_inventory(inventory, Some(version))
+        ObjectVersionDetails::from_inventory(inventory, version.into())
     }
 
     /// Returns a vector containing the version metadata for ever version of an object. The vector
@@ -287,8 +321,8 @@ impl OcflRepo {
     pub fn get_object_file(
         &self,
         object_id: &str,
-        path: &InventoryPath,
-        version_num: Option<VersionNum>,
+        path: &LogicalPath,
+        version_num: VersionRef,
         sink: &mut dyn Write,
     ) -> Result<()> {
         self.ensure_open()?;
@@ -303,24 +337,24 @@ impl OcflRepo {
     pub fn get_staged_object_file(
         &self,
         object_id: &str,
-        path: &InventoryPath,
+        path: &LogicalPath,
         sink: &mut dyn Write,
     ) -> Result<()> {
         self.ensure_open()?;
 
         let inventory = self.get_staged_inventory(object_id)?;
-        let content_path = inventory.content_path_for_logical_path(path, None)?;
+        let content_path = inventory.content_path_for_logical_path(path, VersionRef::Head)?;
 
         let version_prefix = format!("{}/", inventory.head);
 
-        if content_path.as_ref().as_ref().starts_with(&version_prefix) {
+        if content_path.starts_with(&version_prefix) {
             // The content exists in staging
             self.get_staging()?
-                .get_object_file(object_id, path, None, sink)
+                .get_object_file(object_id, path, VersionRef::Head, sink)
         } else {
             // The content exists in the main repo
             self.store
-                .get_object_file(object_id, path, Some(inventory.head.previous()?), sink)
+                .get_object_file(object_id, path, inventory.head.previous()?.into(), sink)
         }
     }
 
@@ -331,7 +365,7 @@ impl OcflRepo {
     pub fn list_file_versions(
         &self,
         object_id: &str,
-        path: &InventoryPath,
+        path: &LogicalPath,
     ) -> Result<Vec<VersionDetails>> {
         self.ensure_open()?;
 
@@ -342,7 +376,7 @@ impl OcflRepo {
         let mut current_digest: Option<Rc<HexDigest>> = None;
 
         for (id, version) in inventory.versions {
-            match version.lookup_digest(&path) {
+            match version.lookup_digest(path) {
                 Some(digest) => {
                     if current_digest.is_none()
                         || current_digest.as_ref().unwrap().as_ref().ne(digest)
@@ -395,7 +429,7 @@ impl OcflRepo {
             return Ok(Vec::new());
         }
 
-        match self.get_staging()?.get_inventory(&object_id) {
+        match self.get_staging()?.get_inventory(object_id) {
             Err(RocflError::NotFound(_)) => Ok(Vec::new()),
             Err(e) => Err(e),
             Ok(inventory) => inventory.diff_versions(None, inventory.head),
@@ -432,30 +466,13 @@ impl OcflRepo {
 
         let object_id = object_id.trim();
 
-        if object_id.is_empty() {
-            return Err(RocflError::IllegalArgument(
-                "Object IDs may not be blank".to_string(),
-            ));
-        }
-
-        if digest_algorithm != DigestAlgorithm::Sha512
-            && digest_algorithm != DigestAlgorithm::Sha256
-        {
-            return Err(RocflError::IllegalArgument(format!(
-                "The inventory digest algorithm must be sha512 or sha256. Found: {}",
-                digest_algorithm.to_string()
-            )));
-        }
-
-        if content_dir.eq(".") || content_dir.eq("..") || content_dir.contains('/') {
-            return Err(RocflError::IllegalArgument(
-                format!("The content directory cannot equal '.' or '..' and cannot contain a '/'. Found: {}",
-                        content_dir)));
-        }
+        validate::validate_object_id(object_id)?;
+        validate::validate_digest_algorithm(digest_algorithm)?;
+        validate::validate_content_dir(content_dir)?;
 
         let _lock = self.get_lock_manager()?.acquire(object_id)?;
 
-        match self.store.get_inventory(&object_id) {
+        match self.store.get_inventory(object_id) {
             Err(RocflError::NotFound(_)) => (),
             Err(e) => return Err(e),
             _ => {
@@ -466,7 +483,7 @@ impl OcflRepo {
             }
         }
 
-        let version_num = VersionNum::with_width(1, padding_width);
+        let version_num = VersionNum::v1_with_width(padding_width);
 
         let mut inventory = Inventory::builder(object_id)
             .with_digest_algorithm(digest_algorithm)
@@ -501,7 +518,7 @@ impl OcflRepo {
     pub fn copy_files_internal(
         &self,
         object_id: &str,
-        version_num: Option<VersionNum>,
+        version_num: VersionRef,
         src: &[impl AsRef<str>],
         dst: &str,
         recursive: bool,
@@ -515,7 +532,7 @@ impl OcflRepo {
         let _lock = self.get_lock_manager()?.acquire(object_id)?;
 
         let mut inventory = self.get_or_created_staged_inventory(object_id)?;
-        let src_version_num = version_num.unwrap_or(inventory.head);
+        let src_version_num = version_num.resolve(inventory.head);
         let staging = self.get_staging()?;
 
         let (to_copy, mut errors) =
@@ -729,7 +746,7 @@ impl OcflRepo {
 
         let _lock = self.get_lock_manager()?.acquire(object_id)?;
 
-        let mut inventory = match staging.get_inventory(&object_id) {
+        let mut inventory = match staging.get_inventory(object_id) {
             Ok(inventory) => inventory,
             Err(RocflError::NotFound(_)) => return Ok(()),
             Err(e) => return Err(e),
@@ -759,7 +776,7 @@ impl OcflRepo {
         let reset_adds = head_paths
             .into_iter()
             .filter(|path| !previous_paths.contains(path))
-            .collect::<HashSet<Rc<InventoryPath>>>();
+            .collect::<HashSet<Rc<LogicalPath>>>();
 
         // Need to apply add resets first to attempt to avoid path conflicts
         for path in reset_adds {
@@ -806,7 +823,7 @@ impl OcflRepo {
 
         let _lock = self.get_lock_manager()?.acquire(object_id)?;
 
-        let mut inventory = match staging.get_inventory(&object_id) {
+        let mut inventory = match staging.get_inventory(object_id) {
             Ok(inventory) => inventory,
             Err(RocflError::NotFound(_)) => {
                 return Err(RocflError::General(format!(
@@ -829,7 +846,7 @@ impl OcflRepo {
             &duplicates
                 .iter()
                 .map(|p| p.as_ref())
-                .collect::<Vec<&InventoryPath>>(),
+                .collect::<Vec<&ContentPath>>(),
         )?;
         staging.rm_orphaned_files(&inventory)?;
 
@@ -857,10 +874,10 @@ impl OcflRepo {
     fn get_or_created_staged_inventory(&self, object_id: &str) -> Result<Inventory> {
         let staging = self.get_staging()?;
 
-        match staging.get_inventory(&object_id) {
+        match staging.get_inventory(object_id) {
             Ok(inventory) => Ok(inventory),
             Err(RocflError::NotFound(_)) => {
-                let mut inventory = self.store.get_inventory(&object_id)?;
+                let mut inventory = self.store.get_inventory(object_id)?;
 
                 if inventory.mutable_head {
                     return Err(RocflError::IllegalState(
@@ -912,7 +929,7 @@ impl OcflRepo {
         src: &[impl AsRef<Path>],
         dst: &str,
         recursive: bool,
-        operator: impl Fn(&Path, InventoryPath, &mut Inventory) -> Result<()>,
+        operator: impl Fn(&Path, LogicalPath, &mut Inventory) -> Result<()>,
     ) -> Result<()> {
         if src.is_empty() {
             return Ok(());
@@ -1023,7 +1040,7 @@ impl OcflRepo {
     fn copy_file(
         &self,
         file: impl AsRef<Path>,
-        logical_path: InventoryPath,
+        logical_path: LogicalPath,
         inventory: &mut Inventory,
     ) -> Result<()> {
         let mut reader = inventory.digest_algorithm.reader(File::open(&file)?)?;
@@ -1038,7 +1055,7 @@ impl OcflRepo {
         // paths were already validated for conflicts. It is possible the file move could fail
         // if the source files conflict, but this will not corrupt anything.
         self.get_staging()?
-            .stage_file_copy(&inventory, &mut reader, &logical_path)?;
+            .stage_file_copy(inventory, &mut reader, &logical_path)?;
         let digest = reader.finalize_hex();
         inventory.add_file_to_head(digest, logical_path)
     }
@@ -1046,7 +1063,7 @@ impl OcflRepo {
     fn move_file(
         &self,
         file: impl AsRef<Path>,
-        logical_path: InventoryPath,
+        logical_path: LogicalPath,
         inventory: &mut Inventory,
     ) -> Result<()> {
         info!(
@@ -1063,7 +1080,7 @@ impl OcflRepo {
         // paths were already validated for conflicts. It is possible the file move could fail
         // if the source files conflict, but this will not corrupt anything.
         self.get_staging()?
-            .stage_file_move(&inventory, &file, &logical_path)?;
+            .stage_file_move(inventory, &file, &logical_path)?;
         inventory.add_file_to_head(digest, logical_path)
     }
 
@@ -1077,7 +1094,7 @@ impl OcflRepo {
         src: &[impl AsRef<str>],
         dst: &str,
         recursive: bool,
-    ) -> Result<(HashMap<Rc<InventoryPath>, InventoryPath>, Vec<String>)> {
+    ) -> Result<(HashMap<Rc<LogicalPath>, LogicalPath>, Vec<String>)> {
         let mut to_move = HashMap::new();
         let mut errors = Vec::new();
 
@@ -1124,7 +1141,7 @@ impl OcflRepo {
                             {
                                 logical_path_in_dst_dir_internal(&file, &dir.parent(), dst)?
                             } else {
-                                logical_path_in_dst_dir_internal(&file, &dir, dst)?
+                                logical_path_in_dst_dir_internal(&file, dir, dst)?
                             };
 
                             has_matches = true;
@@ -1216,7 +1233,7 @@ impl OcflRepo {
 
 /// An iterator that adapts the output of a delegate `Inventory` iterator into another type.
 struct InventoryAdapterIter<'a, T> {
-    iter: Box<dyn Iterator<Item = Inventory> + 'a>,
+    iter: Box<dyn Iterator<Item = Result<Inventory>> + 'a>,
     adapter: Box<dyn Fn(Inventory) -> Result<T>>,
 }
 
@@ -1224,7 +1241,7 @@ impl<'a, T> InventoryAdapterIter<'a, T> {
     /// Creates a new `InventoryAdapterIter` that applies the `adapter` closure to the output
     /// of every `next()` call.
     fn new(
-        iter: Box<dyn Iterator<Item = Inventory> + 'a>,
+        iter: Box<dyn Iterator<Item = Result<Inventory>> + 'a>,
         adapter: impl Fn(Inventory) -> Result<T> + 'a + 'static,
     ) -> Self {
         Self {
@@ -1235,18 +1252,16 @@ impl<'a, T> InventoryAdapterIter<'a, T> {
 }
 
 impl<'a, T> Iterator for InventoryAdapterIter<'a, T> {
-    type Item = T;
+    type Item = Result<T>;
 
     fn next(&mut self) -> Option<Self::Item> {
         match self.iter.next() {
             None => None,
-            Some(inventory) => match self.adapter.deref()(inventory) {
-                Ok(adapted) => Some(adapted),
-                Err(e) => {
-                    error!("{:#}", e);
-                    self.next()
-                }
+            Some(Ok(inventory)) => match self.adapter.deref()(inventory) {
+                Ok(adapted) => Some(Ok(adapted)),
+                Err(e) => Some(Err(e)),
             },
+            Some(Err(e)) => Some(Err(e)),
         }
     }
 }
@@ -1256,7 +1271,7 @@ fn logical_path_in_dst_dir(
     src: impl AsRef<Path>,
     base: impl AsRef<Path>,
     dst: &str,
-) -> Result<InventoryPath> {
+) -> Result<LogicalPath> {
     let mut logical_path = dst.to_string();
     if !logical_path.ends_with('/') {
         logical_path.push('/');
@@ -1269,24 +1284,24 @@ fn logical_path_in_dst_dir(
     logical_path.try_into()
 }
 
-/// Same as `logical_path_in_dst_dir()` but operates on `InventoryPath`s
+/// Same as `logical_path_in_dst_dir()` but operates on `LogicalPath`s
 fn logical_path_in_dst_dir_internal(
-    src: &InventoryPath,
-    base: &InventoryPath,
+    src: &LogicalPath,
+    base: &LogicalPath,
     dst: &str,
-) -> Result<InventoryPath> {
+) -> Result<LogicalPath> {
     let mut logical_path = dst.to_string();
     if !logical_path.ends_with('/') {
         logical_path.push('/');
     }
 
-    let base_length = if base.as_ref().is_empty() {
+    let base_length = if base.as_str().is_empty() {
         0
     } else {
-        base.as_ref().len() + 1
+        base.as_str().len() + 1
     };
 
-    logical_path.push_str(&src.as_ref().as_str()[base_length..]);
+    logical_path.push_str(&src.as_str()[base_length..]);
     logical_path.try_into()
 }
 
@@ -1297,18 +1312,19 @@ fn logical_path_in_dst_dir_internal(
 fn lookup_staged_digest_and_content_path(
     inventory: &Inventory,
     src_version_num: VersionNum,
-    src_path: &InventoryPath,
-) -> Result<Option<(HexDigest, Rc<InventoryPath>)>> {
+    src_path: &LogicalPath,
+) -> Result<Option<(HexDigest, Rc<ContentPath>)>> {
     let staging_prefix = format!("{}/", inventory.head);
 
     match inventory
         .get_version(src_version_num)?
-        .lookup_digest(&src_path)
+        .lookup_digest(src_path)
     {
         Some(digest) => {
-            let content_path = inventory.content_path_for_digest(&digest, None, Some(&src_path))?;
+            let content_path =
+                inventory.content_path_for_digest(digest, VersionRef::Head, Some(src_path))?;
 
-            if content_path.as_ref().as_ref().starts_with(&staging_prefix) {
+            if content_path.starts_with(&staging_prefix) {
                 Ok(Some((digest.as_ref().clone(), content_path.clone())))
             } else {
                 Ok(None)
