@@ -23,6 +23,7 @@ use rusoto_s3::{
     S3Client as RusotoS3Client, StreamingBody, UploadPartRequest, S3,
 };
 use tokio::io::AsyncReadExt;
+use tokio::runtime;
 use tokio::runtime::Runtime;
 use walkdir::WalkDir;
 
@@ -31,7 +32,12 @@ use super::{OcflLayout, OcflStore};
 use crate::ocfl::consts::*;
 use crate::ocfl::error::{not_found, Result, RocflError};
 use crate::ocfl::inventory::Inventory;
-use crate::ocfl::{paths, specs, util, InventoryPath, LayoutExtensionName, VersionNum};
+use crate::ocfl::paths::{join, join_with_trailing_slash};
+use crate::ocfl::store::{Listing, Storage};
+use crate::ocfl::validate::{IncrementalValidator, ObjectValidationResult, Validator};
+use crate::ocfl::{
+    paths, specs, util, InventoryPath, LayoutExtensionName, LogicalPath, VersionRef,
+};
 
 const TYPE_PLAIN: &str = "text/plain; charset=UTF-8";
 const TYPE_MARKDOWN: &str = "text/markdown; charset=UTF-8";
@@ -43,9 +49,10 @@ const PART_SIZE: u64 = 1024 * 1024 * 5;
 static EXTENSIONS_DIR_SUFFIX: Lazy<String> = Lazy::new(|| format!("/{}", EXTENSIONS_DIR));
 
 pub struct S3OcflStore {
-    s3_client: S3Client,
+    s3_client: Arc<S3Client>,
     /// Maps object IDs to paths within the storage root
     storage_layout: Option<StorageLayout>,
+    validator: Validator<S3Storage>,
     // TODO this never expires entries and is only intended to be useful within the scope of the cli
     /// Caches object ID to path mappings
     id_path_cache: RwLock<HashMap<String, String>>,
@@ -66,7 +73,10 @@ impl S3OcflStore {
         check_extensions(&s3_client);
         let storage_layout = load_storage_layout(&s3_client);
 
+        let s3_client = Arc::new(s3_client);
+
         Ok(Self {
+            validator: Validator::new(S3Storage::new(s3_client.clone())),
             s3_client,
             storage_layout,
             id_path_cache: RwLock::new(HashMap::new()),
@@ -87,7 +97,10 @@ impl S3OcflStore {
 
         init_new_repo(&s3_client, layout.as_ref())?;
 
+        let s3_client = Arc::new(s3_client);
+
         Ok(Self {
+            validator: Validator::new(S3Storage::new(s3_client.clone())),
             s3_client,
             storage_layout: layout,
             id_path_cache: RwLock::new(HashMap::new()),
@@ -136,21 +149,17 @@ impl S3OcflStore {
             &object_id
         );
 
-        let mut iter = InventoryIter::new_id_matching(&self, &object_id, self.closed.clone());
+        let mut iter = InventoryIter::new_id_matching(self, object_id, self.closed.clone());
 
-        match iter.next() {
-            Some(inventory) => Ok(inventory),
-            None => Err(not_found(&object_id, None)),
-        }
-    }
-
-    fn parse_inventory_optional(&self, object_root: &str) -> Option<Inventory> {
-        match self.parse_inventory(object_root) {
-            Ok(inventory) => inventory,
-            Err(e) => {
-                error!("{:#}", e);
-                None
-            }
+        loop {
+            return match iter.next() {
+                Some(Ok(inventory)) => Ok(inventory),
+                Some(Err(e)) => {
+                    error!("{:#}", e);
+                    continue;
+                }
+                None => Err(not_found(object_id, None)),
+            };
         }
     }
 
@@ -177,7 +186,7 @@ impl S3OcflStore {
     /// This is normally the `inventory.json` file in the object's root, but it could also be
     /// the inventory file in an extension directory, such as the mutable HEAD extension.
     fn parse_inventory(&self, object_root: &str) -> Result<Option<Inventory>> {
-        let bytes = self.get_inventory_bytes(&object_root)?;
+        let bytes = self.get_inventory_bytes(object_root)?;
         // TODO should validate hash
 
         if let Some((bytes, mutable_head)) = bytes {
@@ -206,7 +215,6 @@ impl S3OcflStore {
 
     fn parse_inventory_bytes(&self, bytes: &[u8]) -> Result<Inventory> {
         let inventory: Inventory = serde_json::from_slice(bytes)?;
-        inventory.validate()?;
         Ok(inventory)
     }
 
@@ -289,7 +297,7 @@ impl S3OcflStore {
         mut callable: impl FnMut(&mut Vec<String>) -> Result<()>,
     ) -> Result<Vec<String>> {
         if let Err(e) = callable(&mut done) {
-            for path in done.iter() {
+            for path in &done {
                 if let Err(e2) = self.s3_client.delete_object(path) {
                     error!("Failed to rollback file {}: {}", path, e2);
                 }
@@ -327,7 +335,7 @@ impl OcflStore for S3OcflStore {
 
         match self.get_object_root_path(object_id) {
             Some(object_root) => self.parse_inventory_required(object_id, &object_root),
-            None => self.scan_for_inventory(&object_id),
+            None => self.scan_for_inventory(object_id),
         }
     }
 
@@ -337,12 +345,12 @@ impl OcflStore for S3OcflStore {
     fn iter_inventories<'a>(
         &'a self,
         filter_glob: Option<&str>,
-    ) -> Result<Box<dyn Iterator<Item = Inventory> + 'a>> {
+    ) -> Result<Box<dyn Iterator<Item = Result<Inventory>> + 'a>> {
         self.ensure_open()?;
 
         Ok(Box::new(match filter_glob {
-            Some(glob) => InventoryIter::new_glob_matching(&self, glob, self.closed.clone())?,
-            None => InventoryIter::new(&self, None, self.closed.clone()),
+            Some(glob) => InventoryIter::new_glob_matching(self, glob, self.closed.clone())?,
+            None => InventoryIter::new(self, None, self.closed.clone()),
         }))
     }
 
@@ -352,8 +360,8 @@ impl OcflStore for S3OcflStore {
     fn get_object_file(
         &self,
         object_id: &str,
-        path: &InventoryPath,
-        version_num: Option<VersionNum>,
+        path: &LogicalPath,
+        version_num: VersionRef,
         sink: &mut dyn Write,
     ) -> Result<()> {
         self.ensure_open()?;
@@ -361,7 +369,7 @@ impl OcflStore for S3OcflStore {
         let inventory = self.get_inventory(object_id)?;
 
         let content_path = inventory.content_path_for_logical_path(path, version_num)?;
-        let storage_path = join(&inventory.object_root, content_path.as_ref().as_ref());
+        let storage_path = join(&inventory.object_root, content_path.as_str());
 
         self.s3_client.stream_object(&storage_path, sink)
     }
@@ -519,10 +527,48 @@ impl OcflStore for S3OcflStore {
         Ok(extensions)
     }
 
+    /// Validates the specified object and returns any problems found. Err will only be returned
+    /// if a non-validation problem was encountered.
+    fn validate_object(
+        &self,
+        object_id: &str,
+        fixity_check: bool,
+    ) -> Result<ObjectValidationResult> {
+        let object_root = self.lookup_or_find_object_root_path(object_id)?;
+
+        self.validator
+            .validate_object(Some(object_id), &object_root, fixity_check)
+    }
+
+    /// Validates the specified object at the specified path, relative the storage root, and
+    /// returns any problems found. Err will only be returned if a non-validation problem was
+    /// encountered.
+    fn validate_object_at(
+        &self,
+        object_root: &str,
+        fixity_check: bool,
+    ) -> Result<ObjectValidationResult> {
+        self.validator
+            .validate_object(None, object_root, fixity_check)
+    }
+
+    /// Validates the structure of an OCFL repository as well as all of the objects in the repository
+    /// When `fixity_check` is `false`, then the digests of object content files are not validated.
+    ///
+    /// The storage root is validated immediately, and an incremental validator is returned that
+    /// is used to lazily validate the rest of the repository.
+    fn validate_repo<'a>(
+        &'a self,
+        fixity_check: bool,
+    ) -> Result<Box<dyn IncrementalValidator + 'a>> {
+        Ok(Box::new(self.validator.validate_repo(fixity_check)?))
+    }
+
     /// Instructs the store to gracefully stop any in-flight work and not accept any additional
     /// requests.
     fn close(&self) {
         self.closed.store(true, Ordering::Release);
+        self.validator.close();
     }
 }
 
@@ -530,6 +576,7 @@ struct S3Client {
     s3_client: RusotoS3Client,
     bucket: String,
     prefix: String,
+    // TODO this should ideally be externalized, but wait for new aws rust client
     runtime: Runtime,
 }
 
@@ -546,6 +593,10 @@ struct InventoryIter<'a> {
     closed: Arc<AtomicBool>,
 }
 
+pub struct S3Storage {
+    s3_client: Arc<S3Client>,
+}
+
 impl S3Client {
     fn new(
         region: Region,
@@ -557,7 +608,7 @@ impl S3Client {
             s3_client: create_rusoto_client(region, profile),
             bucket: bucket.to_owned(),
             prefix: prefix.unwrap_or_default().to_owned(),
-            runtime: Runtime::new()?,
+            runtime: runtime::Builder::new_multi_thread().enable_all().build()?,
         })
     }
 
@@ -576,7 +627,7 @@ impl S3Client {
     /// Returns all of the object keys or logical directories that are under the specified prefix.
     /// All returned keys and key parts are relative the repository prefix; not the search prefix.
     fn list_prefix(&self, path: &str, delimiter: Option<String>) -> Result<ListResult> {
-        let prefix = join_with_trailing_slash(&self.prefix, &path);
+        let prefix = join_with_trailing_slash(&self.prefix, path);
 
         info!("Listing S3 prefix: {}", prefix);
 
@@ -629,7 +680,7 @@ impl S3Client {
     }
 
     fn get_object(&self, path: &str) -> Result<Option<Vec<u8>>> {
-        let key = join(&self.prefix, &path);
+        let key = join(&self.prefix, path);
 
         info!("Getting object from S3: {}", key);
 
@@ -658,7 +709,7 @@ impl S3Client {
     }
 
     fn stream_object(&self, path: &str, sink: &mut dyn Write) -> Result<()> {
-        let key = join(&self.prefix, &path);
+        let key = join(&self.prefix, path);
 
         info!("Streaming object from S3: {}", key);
 
@@ -688,7 +739,7 @@ impl S3Client {
     }
 
     fn delete_object(&self, path: &str) -> Result<()> {
-        let key = join(&self.prefix, &path);
+        let key = join(&self.prefix, path);
 
         info!("Deleting object in S3: {}", key);
 
@@ -708,7 +759,7 @@ impl S3Client {
         content: Bytes,
         content_type: Option<&str>,
     ) -> Result<()> {
-        let key = join(&self.prefix, &path);
+        let key = join(&self.prefix, path);
 
         info!("Putting object in S3: {}", key);
 
@@ -738,7 +789,7 @@ impl S3Client {
         if content_length > PART_SIZE {
             self.multipart_put_file(path, file_path, content_length, content_type)?;
         } else {
-            let key = join(&self.prefix, &path);
+            let key = join(&self.prefix, path);
             info!(
                 "Putting {} in S3 at {}",
                 file_path.as_ref().to_string_lossy(),
@@ -770,7 +821,7 @@ impl S3Client {
         content_length: u64,
         content_type: Option<&str>,
     ) -> Result<()> {
-        let key = join(&self.prefix, &path);
+        let key = join(&self.prefix, path);
 
         info!(
             "Initiating S3 multipart upload of {} to {}",
@@ -921,32 +972,30 @@ impl<'a> InventoryIter<'a> {
         }
     }
 
-    fn create_if_matches(&self, object_root: &str) -> Option<Inventory> {
-        match self.store.parse_inventory_optional(object_root) {
-            Some(inventory) => {
+    fn create_if_matches(&self, object_root: &str) -> Option<Result<Inventory>> {
+        match self.store.parse_inventory(object_root) {
+            Ok(Some(inventory)) => {
                 if let Some(id_matcher) = &self.id_matcher {
                     if id_matcher(&inventory.id) {
-                        Some(inventory)
+                        Some(Ok(inventory))
                     } else {
                         None
                     }
                 } else {
-                    Some(inventory)
+                    Some(Ok(inventory))
                 }
             }
-            None => {
-                error!(
-                    "Expected object to exist at {}, but none found.",
-                    object_root
-                );
-                None
-            }
+            Ok(None) => Some(Err(RocflError::NotFound(format!(
+                "Expected object to exist at {}, but none found.",
+                object_root
+            )))),
+            Err(e) => Some(Err(e)),
         }
     }
 }
 
 impl<'a> Iterator for InventoryIter<'a> {
-    type Item = Inventory;
+    type Item = Result<Inventory>;
 
     fn next(&mut self) -> Option<Self::Item> {
         loop {
@@ -972,18 +1021,20 @@ impl<'a> Iterator for InventoryIter<'a> {
                         continue;
                     }
 
-                    match self.store.list_dir(&entry) {
-                        Ok(listing) => {
-                            if is_object_dir(&listing.objects) {
-                                if let Some(inventory) = self.create_if_matches(&entry) {
-                                    return Some(inventory);
-                                }
-                            } else {
-                                self.dir_iters.push(self.current.replace(None).unwrap());
-                                self.current.replace(Some(listing.directories.into_iter()));
-                            }
+                    let listing = match self.store.list_dir(&entry) {
+                        Ok(listing) => listing,
+                        Err(e) => return Some(Err(e)),
+                    };
+
+                    if is_object_dir(&listing.objects) {
+                        match self.create_if_matches(&entry) {
+                            Some(Ok(inventory)) => return Some(Ok(inventory)),
+                            Some(Err(e)) => return Some(Err(e)),
+                            _ => (),
                         }
-                        Err(e) => error!("{:#}", e),
+                    } else {
+                        self.dir_iters.push(self.current.replace(None).unwrap());
+                        self.current.replace(Some(listing.directories.into_iter()));
                     }
                 }
             }
@@ -991,6 +1042,55 @@ impl<'a> Iterator for InventoryIter<'a> {
     }
 }
 
+impl S3Storage {
+    fn new(s3_client: Arc<S3Client>) -> Self {
+        Self { s3_client }
+    }
+}
+
+impl Storage for S3Storage {
+    /// Reads the file at the specified path and writes its contents to the provided sink.
+    fn read<W: Write>(&self, path: &str, sink: &mut W) -> Result<()> {
+        self.s3_client.stream_object(path, sink)
+    }
+
+    /// Lists the contents of the specified directory. If `recursive` is `true`, then all leaf-nodes
+    /// are returned. If the directory does not exist, or is empty, then an empty vector is returned.
+    /// The returned paths are all relative the directory that was listed.
+    fn list(&self, path: &str, recursive: bool) -> Result<Vec<Listing>> {
+        let prefix_len = if path.is_empty() || path.ends_with('/') {
+            path.len()
+        } else {
+            path.len() + 1
+        };
+
+        if recursive {
+            let key_parts = self.s3_client.list_objects(path)?;
+            Ok(key_parts
+                .iter()
+                .map(|entry| Listing::file_owned(entry[prefix_len..].to_string()))
+                .collect::<Vec<Listing>>())
+        } else {
+            let s3_result = self.s3_client.list_dir(path)?;
+            let mut result =
+                Vec::with_capacity(s3_result.directories.len() + s3_result.objects.len());
+
+            s3_result
+                .objects
+                .iter()
+                .map(|entry| Listing::file_owned(entry[prefix_len..].to_string()))
+                .for_each(|entry| result.push(entry));
+            s3_result
+                .directories
+                .iter()
+                .map(|entry| Listing::dir_owned(entry[prefix_len..].to_string()))
+                .for_each(|entry| result.push(entry));
+
+            Ok(result)
+        }
+    }
+}
+
 fn check_extensions(s3_client: &S3Client) {
     match s3_client.list_dir(EXTENSIONS_DIR) {
         Ok(result) => {
@@ -1169,32 +1269,6 @@ fn is_object_dir(objects: &[String]) -> bool {
     false
 }
 
-fn join(part1: &str, part2: &str) -> String {
-    let mut joined = match part1.ends_with('/') {
-        true => part1[..part1.len() - 1].to_string(),
-        false => part1.to_string(),
-    };
-
-    if !part2.is_empty() {
-        if (!joined.is_empty() || part1 == "/") && !part2.starts_with('/') {
-            joined.push('/');
-        }
-        joined.push_str(part2);
-    }
-
-    joined
-}
-
-fn join_with_trailing_slash(part1: &str, part2: &str) -> String {
-    let mut joined = join(part1, part2);
-
-    if !joined.is_empty() && !joined.ends_with('/') {
-        joined.push('/');
-    }
-
-    joined
-}
-
 #[cfg(test)]
 mod tests {
     use super::{is_object_dir, join, join_with_trailing_slash};
